import numpy as np
from CyRK import pysolve_ivp

def diffeq(dy, t, y,):
    # Real dy
    dy[0] = 3.1 * t - y[1]
    dy[1] = y[0] * (0.3 * t * y[1])
    # Extra output
    dy[2] = 0.25
    dy[3] = t / 2.

t_span = (0.0, 10.0)
y0 = np.asarray([5., 2.], dtype=np.float64)

def test_pysolve_extra_output_with_dense():
    """ Test that pysolve (and by extension, cysolve) can interpolate extra outputs when `dense_output = True` """

    result = pysolve_ivp(
        diffeq,
        t_span,
        y0,
        method = 'RK45',
        t_eval = None,
        dense_output = True,
        args = None,
        expected_size = 0,
        num_extra = 2,
        first_step = 0.0,
        max_step = 100_000,
        rtol = 1.0e-3,
        atol = 1.0e-6,
        max_num_steps = 0,
        max_ram_MB = 2000,
        pass_dy_as_arg = True
        )

    assert result.success

    # Call dense output with a float
    dense_out_float = result(0.3)
    expected_float = np.asarray([[4.52627329], [2.13093483], [0.25      ], [0.15      ]], dtype=np.float64)
    assert dense_out_float.shape == (4, 1)
    assert np.allclose(dense_out_float, expected_float)

    # Call dense output with an array
    dense_out_array = result(np.asarray([0.3, 4.0, 8., 9.10, 9.9]))
    expected_array = np.asarray([[ 4.52627329,  1.87160845,  0.13355624,  1.57871591, -1.01590877],
                                 [ 2.13093483, 21.05384256, 10.31201731, 48.55169176, 57.06848508],
                                 [ 0.25,        0.25,        0.25,        0.25,        0.25      ],
                                 [ 0.15,        2.,          4.,          4.55,        4.95      ]], dtype=np.float64)
    assert dense_out_array.shape == (4, 5)
    assert np.allclose(dense_out_array, expected_array)



# Find Version Number
import importlib.metadata
__version__ = importlib.metadata.version("CyRK")
version = __version__

# Import numba solver
from .nb.nbrk import nbsolve_ivp

# Import helper functions
from .helper import nb2cy, cy2nb

# Import test functions
from ._test import test_nbrk, test_cysolver, test_pysolver

# Import python solver
from CyRK.cy.cysolver_api import WrapCySolverResult
from CyRK.cy.pysolver import pysolve_ivp

# Helper function that provides directories to CyRK c++ headers
def get_include():
    import os
    import CyRK
    cyrk_dir = os.path.dirname(CyRK.__file__)

    cyrk_dirs = list()
    cyrk_dirs.append(
        os.path.join(cyrk_dir, 'cy')  # CySolver headers
    )
    cyrk_dirs.append(
        os.path.join(cyrk_dir, 'array')  # Array headers
    )

    return cyrk_dirs



""" Pendulum Differential Equations """

import numpy as np
from numba import njit

from CyRK import nb2cy

pendulum_y0 = np.asarray((0.01, 0.), dtype=np.float64)
pendulum_args = (1., 1., 9.81)  # length [m], mass [kg], acceleration due to gravity [m s-2]
pendulum_time_span_1 = (0., 10.)
pendulum_time_span_2 = (0., 100.)


@njit(cache=True)
def pendulum_nb(t, y, l, m, g):

    # External torque
    torque = 0.1 * np.sin(t)

    y0 = y[0]  # Angular deflection [rad]
    y1 = y[1]  # Angular velocity [rad s-1]
    dy = np.empty_like(y)
    dy[0] = y1
    dy[1] = (-3. * g / (2. * l)) * np.sin(y0) + (3. / (m * l**2)) * torque
    return dy


pendulum_cy = nb2cy(pendulum_nb, use_njit=True, cache_njit=True)



import os.path

import timeit
from datetime import datetime

import numpy as np
import time
from lotkavolterra import (
    lotkavolterra_cy, lotkavolterra_nb, lotkavolterra_args, lotkavolterra_y0,
    lotkavolterra_time_span_1, lotkavolterra_time_span_2)

from pendulum import pendulum_cy, pendulum_nb, pendulum_args, pendulum_y0, pendulum_time_span_1, pendulum_time_span_2

from lorenz import (
    lorenz_cy, lorenz_nb, lorenz_args, lorenz_y0, lorenz_time_span_1, lorenz_time_span_2,
    lorenz_nb_extra, lorenz_cy_extra)

from CyRK.cy.cysolver_test import cytester

REPEATS = 4
RTOL = 1.e-6
ATOL = 1.e-8

CySolverLotkavolterra_Int = 5
CySolverPendulum_Int = 6
CySolverLorenz_Int = 3
CySolverLorenzExtra_Int = 4

performance_filename = 'cyrk_performance.csv'
diffeqs = {
    'Lotkavolterra'  : (lotkavolterra_cy, lotkavolterra_nb, lotkavolterra_args, lotkavolterra_y0,
                        (lotkavolterra_time_span_1, lotkavolterra_time_span_2), CySolverLotkavolterra_Int),
    'Pendulum'       : (pendulum_cy, pendulum_nb, pendulum_args, pendulum_y0, (pendulum_time_span_1, pendulum_time_span_2), CySolverPendulum_Int),
    'Lorenz'         : (lorenz_cy, lorenz_nb, lorenz_args, lorenz_y0, (lorenz_time_span_1, lorenz_time_span_2), CySolverLorenz_Int),
    'Lorenz-ExtraOut': (lorenz_cy_extra, lorenz_nb_extra, lorenz_args, lorenz_y0,
                        (lorenz_time_span_1, lorenz_time_span_2), CySolverLorenzExtra_Int)
    }

time_spans = {
    'Small Time': 0,
    'Large Time': 1,
    }

statistics = {
    'cython (avg)': 0,
    'cython (std)': 1,
    'CySolver (avg)': 1,
    'CySolver (std)': 2,
    'numba  (avg)': 3,
    'numba  (std)': 4
    }

integration_methods = {
    'RK23'  : 0,
    'RK45'  : 1,
    'DOP853': 2
    }

integration_methods_asstr = {
    0: 'RK23',
    1: 'RK45',
    2: 'DOP853'
    }


def make_performance_file(integration_method_name):

    if integration_method_name not in integration_methods:
        raise ValueError

    performance_filename = f'cyrk_performance-{integration_method_name}.csv'

    # Build Headers
    header_0 = '(all times in ms),'
    header_1 = ','
    header_2 = 'CyRK Version, Run-on Date'

    for d_i, diffeq_name in enumerate(diffeqs):
        for t_i, time_span_name in enumerate(time_spans):
            for s_i, statistics_name in enumerate(statistics):

                if t_i == 0 and s_i == 0:
                    header_0 += f', {diffeq_name}'
                else:
                    header_0 += ','

                if s_i == 0:
                    header_1 += f', {time_span_name}'
                else:
                    header_1 += ','

                header_2 += f', {statistics_name}'

    with open(performance_filename, 'x') as performance_file:
        performance_file.write(header_0 + '\n')
        performance_file.write(header_1 + '\n')
        performance_file.write(header_2 + '\n')

    print('CyRK Performance File Created.')


def run_performance(integration_method_name):

    if integration_method_name not in integration_methods:
        raise ValueError
    int_method = integration_methods[integration_method_name]
    int_method_str = integration_methods_asstr[int_method]
    performance_filename = f'cyrk_performance-{integration_method_name}.csv'

    from CyRK import __version__, pysolve_ivp, nbsolve_ivp
    print(f'Running Performance for CyRK v{__version__} and {integration_method_name}.')
    now = datetime.now()
    dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
    performance_csv_line = f'{__version__}, {dt_string}'
    # Run performance checks
    for d_i, diffeq_name in enumerate(diffeqs):

        print(f'\tWorking on {diffeq_name}')
        cy_diffeq, nb_diffeq, args_, y0, timespans, cysolver_diffeq_int = diffeqs[diffeq_name]
        args_as_array = np.asarray(args_)

        for t_i, time_span_name in enumerate(time_spans):

            print(f'\t\t{time_span_name}')
            t_index = time_spans[time_span_name]
            time_span = timespans[t_index]

            # Run the numba function once to make sure everything is compiled.
            print('\t\tPrecompiling numba')
            _ = nbsolve_ivp(nb_diffeq, time_span, y0, args_, rtol=RTOL, atol=ATOL, rk_method=int_method)

            if 'extraout' in diffeq_name.lower():
                cy_timer = timeit.Timer(lambda: pysolve_ivp(cy_diffeq, time_span, y0, args=args_, rtol=RTOL, atol=ATOL, method=int_method_str, num_extra=3, pass_dy_as_arg=True))
                nb_timer = timeit.Timer(lambda: nbsolve_ivp(nb_diffeq, time_span, y0, args=args_, rtol=RTOL, atol=ATOL, rk_method=int_method, capture_extra=True))
                cysolver_timer = timeit.Timer(lambda: cytester(cysolver_diffeq_int, time_span, y0, args=args_as_array, rtol=RTOL, atol=ATOL, method=int_method))
            else:
                cy_timer = timeit.Timer(lambda: pysolve_ivp(cy_diffeq, time_span, y0, args=args_, rtol=RTOL, atol=ATOL, method=int_method_str, pass_dy_as_arg=True))
                nb_timer = timeit.Timer(lambda: nbsolve_ivp(nb_diffeq, time_span, y0, args=args_, rtol=RTOL, atol=ATOL, rk_method=int_method))
                cysolver_timer = timeit.Timer(lambda: cytester(cysolver_diffeq_int, time_span, y0, args=args_as_array, rtol=RTOL, atol=ATOL, method=int_method))

            # Cython
            print('\t\t\tWorking on pysolver.', end='')
            cython_times = list()
            time_0 = time.time()
            for i in range(REPEATS):
                N, T = cy_timer.autorange()
                cython_times.append(T / N * 1000.)
            print(f' Finished taking {time.time() - time_0:0.1f}s.')
            cython_times = np.asarray(cython_times)
            # Store cython results
            cy_avg = np.average(cython_times)
            cy_std = np.std(cython_times)
            performance_csv_line += f', {cy_avg:0.4f}, {cy_std:0.4f}'

            # Cython Solver Class
            print('\t\t\tWorking on cysolver.', end='')
            cysolver_times = list()
            time_0 = time.time()
            for i in range(REPEATS):
                N, T = cysolver_timer.autorange()
                cysolver_times.append(T / N * 1000.)
            print(f' Finished taking {time.time() - time_0:0.1f}s.')
            cysolver_times = np.asarray(cysolver_times)
            # Store Cython Solver results
            cysolver_avg = np.average(cysolver_times)
            cysolver_std = np.std(cysolver_times)
            performance_csv_line += f', {cysolver_avg:0.4f}, {cysolver_std:0.4f}'

            # Numba
            print('\t\t\tWorking on nbsolve_ivp.', end='')
            numba_times = list()
            time_0 = time.time()
            for i in range(REPEATS):
                N, T = nb_timer.autorange()
                numba_times.append(T / N * 1000.)
            print(f' Finished taking {time.time() - time_0:0.1f}s.')
            numba_times = np.asarray(numba_times)
            # Store numba results
            nb_avg = np.average(numba_times)
            nb_std = np.std(numba_times)
            performance_csv_line += f', {nb_avg:0.4f}, {nb_std:0.4f}'

    # Save results to disk
    if not os.path.isfile(performance_filename):
        print('CyRK Performance File Not Found. Creating...')
        make_performance_file(integration_method_name)

    with open(performance_filename, 'a') as performance_file:
        performance_file.write(performance_csv_line + '\n')


if __name__ == '__main__':

    for integration_method_name in integration_methods:
        performance_filename = f'cyrk_performance-{integration_method_name}.csv'

        # Check if performance file exists. Create it if it does not.
        if not os.path.isfile(performance_filename):
            print('CyRK Performance File Not Found. Creating...')
            make_performance_file(integration_method_name)

        run_performance(integration_method_name)



import numpy as np
import pytest

from CyRK.cy.cysolver_api import WrapCySolverResult
from CyRK.cy.cysolver_test import cytester

args = (0.01, 0.02)

initial_conds = np.asarray((20., 20.), dtype=np.float64, order='C')
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-4
atol = 1.0e-7

rtols = np.asarray((1.0e-4, 1.0e-5), dtype=np.float64, order='C')
atols = np.asarray((1.0e-6, 1.0e-7), dtype=np.float64, order='C')


@pytest.mark.filterwarnings("error")  # Some exceptions get propagated via cython as warnings; we want to make sure the lead to crashes.
@pytest.mark.parametrize('capture_extra', (True, False))
@pytest.mark.parametrize('max_step', (1.0, 100_000.0))
@pytest.mark.parametrize('first_step', (0.0, 0.00001))
@pytest.mark.parametrize('integration_method', (0, 1, 2))
@pytest.mark.parametrize('use_different_tols', (True, False))
@pytest.mark.parametrize('use_rtol_array', (True, False))
@pytest.mark.parametrize('use_atol_array', (True, False))
@pytest.mark.parametrize('use_large_timespan', (True, False))
@pytest.mark.parametrize('use_args', (True, False))
def test_cysolve_ivp(use_args,
                     use_large_timespan, use_atol_array, use_rtol_array, use_different_tols, integration_method,
                     first_step, max_step, capture_extra):
    """Check that the pysolve_ivp function is able to run with various changes to its arguments. """

    if use_atol_array:
        atols_float = np.nan
        atols_array = atols
    else:
        atols_float = atol
        atols_array = None
    if use_rtol_array:
        rtols_float = np.nan
        rtols_array = rtols
    else:
        rtols_float = rtol
        rtols_array = None
    
    if use_different_tols:
        # Check that it can run with smaller tolerances
        if rtols_float is not None:
            rtols_float /= 100.0
        if rtols_array is not None:
            rtols_array /= 100.0
        if atols_float is not None:
            atols_float /= 100.0
        if atols_array is not None:
            atols_array /= 100.0
    
    if use_large_timespan:
        time_span_touse = time_span_large
    else:
        time_span_touse = time_span

    if use_args:
        args_touse = np.asarray(args, dtype=np.float64, order='C')
    else:
        args_touse = None
    
    if capture_extra:
        diffeq_num = 2
    else:
        diffeq_num = 0

    result = \
        cytester(diffeq_num, time_span_touse, initial_conds, args=args_touse,
                 method=integration_method,
                 rtol=rtols_float, atol=atols_float, rtol_array=rtols_array, atol_array=atols_array,
                 max_step=max_step, first_step=first_step)

    assert isinstance(result, WrapCySolverResult)
    assert result.success
    assert result.error_code == 1
    assert result.size > 1
    assert result.message == "Integration completed without issue."
    # Check that the ndarrays make sense
    assert type(result.t) == np.ndarray
    assert result.t.dtype == np.float64
    assert result.y.dtype == np.float64
    assert result.t.size > 1
    assert result.t.size == result.y[0].size
    assert len(result.y.shape) == 2
    assert result.y[0].size == result.y[1].size
    assert result.t.size == result.size
    assert result.y[0].size == result.size

    if capture_extra:
        assert result.y.shape[0] == 4
        assert result.y[2].size == result.y[1].size
        assert result.y[3].size == result.y[1].size
    else:
        assert result.y.shape[0] == 2

    assert type(result.message) == str


@pytest.mark.filterwarnings("error")  # Some exceptions get propagated via cython as warnings; we want to make sure the lead to crashes.
@pytest.mark.parametrize('integration_method', (0, 1, 2))
@pytest.mark.parametrize('t_eval_end', (None, 0.5, 1.0))
@pytest.mark.parametrize('test_dense_output', (False, True))
def test_cysolve_ivp_accuracy(integration_method, t_eval_end, test_dense_output):
    #Check that the cython function solver is able to reproduce a known functions integral with reasonable accuracy

    # TODO: This is only checking one equation. Add other types of diffeqs to provide better coverage.

    def correct_answer(t, c1_, c2_):
        y = np.empty((2, t.size), dtype=np.float64)
        y[0] = -c1_ * np.sin(t) + c2_ * np.cos(t) - (np.cos(t) / 2)  # -c1 * sin(t) + c2 * cos(t) - cos(t) / 2
        # At t=0; y = c2 - 1/2
        y[1] = c2_ * np.sin(t) + c1_ * np.cos(t) + (np.sin(t) / 2)   # c2 * sin(t) + c1 * cos(t) + sin(t) / 2
        # At t=0; x = c1
        return y

    # Initial Conditions
    # y=0 --> c2 = + 1/2
    c2 = 0.5
    # x=1 --> c1 = + 1
    c1 = 1.0
    y0 = np.asarray((0., 1.), dtype=np.float64, order='C')
    time_span_ = (0., 10.)

    # Check if we should test t_eval
    if t_eval_end is None:
        t_eval = None
    else:
        t_eval = np.linspace(0.0, t_eval_end, 50)

    result = \
        cytester(1, time_span_, y0,
                 method=integration_method, dense_output=test_dense_output, t_eval=t_eval,
                 rtol=1.0e-8, atol=1.0e-9,)
    
    # Use the integrator's time domain to build a correct solution
    real_answer = correct_answer(result.t, c1, c2)

    if integration_method == "RK23":
        check_rtol = 1.0e-3
        check_atol = 1.0e-6
    elif integration_method == "DOP853":
        check_rtol = 1.0e-5
        check_atol = 1.0e-8
    else:
        check_rtol = 1.0e-4
        check_atol = 1.0e-7
    
    assert np.allclose(result.y, real_answer, rtol=check_rtol, atol=check_atol)

    if test_dense_output:
        # Check that dense output is working and that it gives decent results
        # Check with a float
        y_array = result(0.5)
        assert type(y_array) == np.ndarray
        assert y_array.shape == (2, 1)

        # Check with array
        t_array = np.linspace(0.1, 0.4, 10)
        y_array = result(t_array)
        assert type(y_array) == np.ndarray
        assert y_array.shape == (2, 10)

        # Check accuracy
        y_array_real = correct_answer(t_array, c1, c2)
        assert np.allclose(y_array, y_array_real, rtol=check_rtol, atol=check_atol)

    # Check the accuracy of the results
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # ax.plot(result.t, result.y[0], 'r', label='PySolver')
    # ax.plot(result.t, result.y[1], 'r:')
    # ax.plot(result.t, real_answer[0], 'b', label='Analytic')
    # ax.plot(result.t, real_answer[1], 'b:')
    # plt.show()


@pytest.mark.filterwarnings("error")  # Some exceptions get propagated via cython as warnings; we want to make sure the lead to crashes.
@pytest.mark.parametrize('cysolve_test_func', (0, 1, 2, 3, 4, 5, 6, 7, 8))
def test_cysolve_ivp_all_diffeqs(cysolve_test_func):
    """ Check all of the currently implemented test diffeqs for cysolve_ivp"""

    result = cytester(cysolve_test_func)

    # There is a weird fail state that occasionally happens on MacOS for Python 3.10 where diffeq 5 (lotkavolterra) fails.
    # See GitHub Issue [#67](https://github.com/jrenaud90/CyRK/issues/67)
    import platform
    on_macos = False
    if platform.system().lower() == 'darwin':
        on_macos = True
    if not result.success and on_macos and cysolve_test_func==5:
        pytest.skip("Weird macos bug on diffeq5")
    
    assert result.success
    assert result.t.size > 0
    assert result.y.size > 0
    assert result.y.shape[0] > 0
    assert np.all(~np.isnan(result.t))
    assert np.all(~np.isnan(result.y))



import numpy as np

import pytest

def test_array_module():
    """ Test that the functions load correctly. """
    from CyRK.array import interpj, interp_complexj, interp, interp_complex, interp_array, interp_complex_array


@pytest.mark.parametrize('t_test', (250.23, 650.0, 0., 1000., 1050., -30., 0.1, 0.5, 0.75, 1.25))
def test_interp(t_test):
    """ Test custom interpolation function (floats). """
    from CyRK.array import interp

    t = np.linspace(0., 1000., 1001, dtype=np.float64)
    x = 100 * np.cos(t)**2

    numpy_interp = np.interp(t_test, t, x)

    cyrk_interp = interp(t_test, t, x)

    assert np.allclose([numpy_interp], [cyrk_interp])

@pytest.mark.parametrize('t_test', (250.23, 650.0, 0., 1000., 1050., -30., 0.1, 0.5, 0.75, 1.25))
def test_interp_complex(t_test):
    """ Test custom interpolation function (complex). """
    from CyRK.array import interp_complex

    t = np.linspace(0., 1000., 1001, dtype=np.float64)
    x = 100 * np.cos(t)**2 - 1.j * np.sin(t)

    numpy_interp = np.interp(t_test, t, x)

    cyrk_interp = interp_complex(t_test, t, x)

    assert np.allclose([numpy_interp], [cyrk_interp])

@pytest.mark.parametrize('t_test', (250.23, 650.0, 0., 1000., 1050., -30., 0.1, 0.5, 0.75, 1.25))
def test_interp_with_provided_j(t_test):
    """ Test custom interpolation function with provided j. """
    from CyRK.array import interpj, interp_complexj, interp, interp_complex

    # Test float function
    t = np.linspace(0., 1000., 1001, dtype=np.float64)
    x = 100 * np.cos(t)**2

    numpy_interp = np.interp(t_test, t, x)

    # Get j
    cyrk_interp_1, provided_j = interpj(t_test, t, x)

    # Use j
    cyrk_interp_2 = interp(t_test, t, x, provided_j=provided_j)

    # Two cyrk versions should be exactly the same.
    assert cyrk_interp_1 == cyrk_interp_2

    assert np.allclose([numpy_interp], [cyrk_interp_1])
    assert np.allclose([numpy_interp], [cyrk_interp_2])

    # Test complex function
    t_cmp = np.linspace(0., 1000., 1001, dtype=np.float64)
    x_cmp = 100 * np.cos(t_cmp)**2 - 1.j * np.sin(t_cmp)
    t_test_cmp = 250.2235

    numpy_interp_cmp = np.interp(t_test_cmp, t_cmp, x_cmp)

    # Get j
    cyrk_interp_1_cmp, provided_j_cmp = interp_complexj(t_test_cmp, t_cmp, x_cmp)
    # Use j
    cyrk_interp_2_cmp = interp_complex(t_test_cmp, t_cmp, x_cmp, provided_j=provided_j_cmp)

    # Two cyrk versions should be exactly the same.
    assert cyrk_interp_1_cmp == cyrk_interp_2_cmp

    assert np.allclose([numpy_interp_cmp], [cyrk_interp_1_cmp])
    assert np.allclose([numpy_interp_cmp], [cyrk_interp_2_cmp])

def test_interp_array():
    """ Test custom interpolation function (float arrays). """
    from CyRK.array import interp_array

    t = np.linspace(0., 1000., 500, dtype=np.float64)
    t_small = np.linspace(0., 1000., 50, dtype=np.float64)
    x = 100 * np.cos(t)**2

    numpy_interp = np.interp(t_small, t, x)

    cyrk_interp = np.empty(t_small.size, dtype=np.float64)
    interp_array(t_small, t, x, cyrk_interp)

    assert np.allclose(numpy_interp, cyrk_interp)

def test_interp_complex_array():
    """ Test custom interpolation function (complex arrays). """
    from CyRK.array import interp_complex_array

    t = np.linspace(0., 1000., 500, dtype=np.float64)
    t_small = np.linspace(0., 1000., 50, dtype=np.float64)
    x = 100 * np.cos(t)**2 - 1.j * np.sin(t)

    numpy_interp = np.interp(t_small, t, x)

    cyrk_interp = np.empty(t_small.size, dtype=np.complex128)
    interp_complex_array(t_small, t, x, cyrk_interp)

    assert np.allclose(numpy_interp, cyrk_interp)



cdef struct LinkedList:
    size_t size
    double* array_ptr
    LinkedList* next

cdef void* allocate_mem(size_t size, char* var_name) noexcept nogil

cdef void* reallocate_mem(void* old_pointer, size_t new_size, char* var_name) noexcept nogil

cdef void free_mem(void* pointer) noexcept nogil


# distutils: language = c
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False
from cython.parallel import prange

import numpy as np

from libc.math cimport isnan, floor

# Get machine precision.
cdef double EPS
EPS = np.finfo(dtype=np.float64).eps


cdef double interp_ptr(
        double desired_x,
        double* x_domain,
        double* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for floats.

    Provided a domain, `x_domain` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `x_domain` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[float]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : float
        Desired value of `dependent_values`.

    """
    # TODO: Needs to be at least 3 item long array. Add exception here?

    cdef double left_value
    left_value = dependent_values[0]
    cdef double right_value
    right_value = dependent_values[len_x - 1]

    # Binary Search with Guess
    cdef Py_ssize_t j
    j = 0
    cdef double slope

    cdef double result
    cdef double fp_at_j
    cdef double xp_at_j
    cdef double fp_at_jp1
    cdef double xp_at_jp1

    # Perform binary search with guess
    if provided_j == -2:
        # No j provided; search for it instead.
        j = c_binary_search_with_guess(desired_x, x_domain, len_x, j)
    elif provided_j < -2:
        # Error
        # TODO: How to handle exception handling in a cdef function... For now just repeat the search.
        j = c_binary_search_with_guess(desired_x, x_domain, len_x, j)
    else:
        j = provided_j

    if j <= -1:
        result = left_value
    elif j >= len_x:
        result = right_value
    else:
        fp_at_j = dependent_values[j]
        xp_at_j = x_domain[j]
        if j == len_x - 1:
            result = fp_at_j
        elif xp_at_j == desired_x:
            result = fp_at_j
        else:
            fp_at_jp1 = dependent_values[j + 1]
            xp_at_jp1 = x_domain[j + 1]
            slope = (fp_at_jp1 - fp_at_j) / (xp_at_jp1 - xp_at_j)

            # If we get nan in one direction, try the other
            result = slope * (desired_x - xp_at_j) + fp_at_j
            if isnan(result):
                result = slope * (desired_x - xp_at_jp1) + fp_at_jp1
                if isnan(result) and (fp_at_jp1 == fp_at_j):
                    result = fp_at_j

    return result


cdef double complex interp_complex_ptr(
        double desired_x,
        double* x_domain,
        double complex* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for complex numbers.

    Provided a domain, `desired_x` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `desired_x` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[complex]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : complex
        Desired value of `dependent_values`.

    """
    # Note: Needs to be at least 3 item long array. Add exception here?

    cdef double complex left_value
    left_value = dependent_values[0]
    cdef double complex right_value
    right_value = dependent_values[len_x - 1]

    # Binary Search with Guess
    cdef Py_ssize_t j
    j = 0
    cdef double slope_real
    cdef double slope_imag
    cdef double x_slope_inverse

    cdef double result_real
    cdef double result_imag
    cdef double complex fp_at_j
    cdef double fp_at_j_real
    cdef double fp_at_j_imag
    cdef double xp_at_j
    cdef double complex fp_at_jp1
    cdef double fp_at_jp1_real
    cdef double fp_at_jp1_imag
    cdef double xp_at_jp1

    # Perform binary search with guess
    if provided_j == -2:
        # No j provided; search for it instead.
        j = c_binary_search_with_guess(desired_x, x_domain, len_x, j)
    elif provided_j < -2:
        # Error
        # TODO: How to handle exception handling in a cdef function... For now just repeat the search.
        j = c_binary_search_with_guess(desired_x, x_domain, len_x, j)
    else:
        j = provided_j

    if j <= -1:
        result_real = left_value.real
        result_imag = left_value.imag
    elif j >= len_x:
        result_real = right_value.real
        result_imag = right_value.imag
    else:
        fp_at_j = dependent_values[j]
        fp_at_j_real = fp_at_j.real
        fp_at_j_imag = fp_at_j.imag
        xp_at_j = x_domain[j]
        if j == len_x - 1:
            result_real = fp_at_j_real
            result_imag = fp_at_j_imag
        elif xp_at_j == desired_x:
            result_real = fp_at_j_real
            result_imag = fp_at_j_imag
        else:
            fp_at_jp1 = dependent_values[j + 1]
            fp_at_jp1_real = fp_at_jp1.real
            fp_at_jp1_imag = fp_at_jp1.imag
            xp_at_jp1 = x_domain[j + 1]
            x_slope_inverse = 1.0 / (xp_at_jp1 - xp_at_j)
            slope_real = (fp_at_jp1_real - fp_at_j_real) * x_slope_inverse
            slope_imag = (fp_at_jp1_imag - fp_at_j_imag) * x_slope_inverse

            # If we get nan in one direction try the other
            # Real Part
            result_real = slope_real * (desired_x - xp_at_j) + fp_at_j_real
            if isnan(result_real):
                result_real = slope_real * (desired_x - xp_at_jp1) + fp_at_jp1_real
                if isnan(result_real) and (fp_at_jp1_real == fp_at_j_real):
                    result_real = fp_at_j_real

            # Imaginary Part
            result_imag = slope_imag * (desired_x - xp_at_j) + fp_at_j_imag
            if isnan(result_imag):
                result_imag = slope_imag * (desired_x - xp_at_jp1) + fp_at_jp1_imag
                if isnan(result_imag) and (fp_at_jp1_imag == fp_at_j_imag):
                    result_imag = fp_at_j_imag

    cdef double complex result
    result = result_real + 1.0j * result_imag

    return result


cpdef double interp(
        double desired_x,
        double[:] x_domain,
        double[:] dependent_values,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for floats.

    Provided a domain, `x_domain` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `x_domain` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[float]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : float
        Desired value of `dependent_values`.

    """

    cdef Py_ssize_t len_x
    len_x = len(x_domain)

    return interp_ptr(desired_x, &x_domain[0], &dependent_values[0], len_x, provided_j)


cpdef double complex interp_complex(
        double desired_x,
        double[:] x_domain,
        double complex[:] dependent_values,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for complex numbers.

    Provided a domain, `desired_x` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `desired_x` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[complex]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : complex
        Desired value of `dependent_values`.

    """

    cdef Py_ssize_t len_x
    len_x = len(x_domain)

    return interp_complex_ptr(desired_x, &x_domain[0], &dependent_values[0], len_x, provided_j)


cdef (double, Py_ssize_t) interpj_ptr(
        double desired_x,
        double* x_domain,
        double* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for floats. This function will return the index that it found during interpolation.

    Provided a domain, `x_domain` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `x_domain` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[float]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : float
        Desired value of `dependent_values`.
    j_out : int
        The index that was found during c_binary_search_with_guess which can be used by other interpolation functions 
        to improve performance.

    """
    # Binary Search with Guess
    cdef Py_ssize_t j_out
    j_out = 0

    # Perform binary search with guess
    if provided_j == -2:
        # No j provided; search for it instead.
        j_out = c_binary_search_with_guess(desired_x, x_domain, len_x, j_out)
    elif provided_j < -2:
        # Error
        # TODO: How to handle exception handling in a cdef function... For now just repeat the search.
        j_out = c_binary_search_with_guess(desired_x, x_domain, len_x, j_out)
    else:
        j_out = provided_j

    return interp_ptr(desired_x, x_domain, dependent_values, len_x, provided_j=j_out), j_out


cpdef (double, Py_ssize_t) interpj(
        double desired_x,
        double[:] x_domain,
        double[:] dependent_values,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for floats. This function will return the index that it found during interpolation.

    Provided a domain, `x_domain` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `x_domain` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[float]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : float
        Desired value of `dependent_values`.
    j_out : int
        The index that was found during c_binary_search_with_guess which can be used by other interpolation functions 
        to improve performance.

    """
    cdef Py_ssize_t len_x
    len_x = len(x_domain)

    return interpj_ptr(desired_x, &x_domain[0], &dependent_values[0], len_x, provided_j)


cdef (double complex, Py_ssize_t) interp_complexj_ptr(
        double desired_x,
        double* x_domain,
        double complex* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for complex numbers.

    Provided a domain, `desired_x` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `desired_x` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[complex]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : complex
        Desired value of `dependent_values`.
    j_out : int
        The index that was found during c_binary_search_with_guess which can be used by other interpolation functions 
        to improve performance.

    """
    # Binary Search with Guess
    cdef Py_ssize_t j_out
    j_out = 0

    # Perform binary search with guess
    if provided_j == -2:
        # No j provided; search for it instead.
        j_out = c_binary_search_with_guess(desired_x, x_domain, len_x, j_out)
    elif provided_j < -2:
        # Error
        # TODO: How to handle exception handling in a cdef function... For now just repeat the search.
        j_out = c_binary_search_with_guess(desired_x, x_domain, len_x, j_out)
    else:
        j_out = provided_j

    return interp_complex_ptr(desired_x, x_domain, dependent_values, len_x, provided_j=j_out), j_out


cpdef (double complex, Py_ssize_t) interp_complexj(
        double desired_x,
        double[:] x_domain,
        double complex[:] dependent_values,
        Py_ssize_t provided_j = -2
        ) noexcept nogil:
    """ Interpolation function for complex numbers.

    Provided a domain, `desired_x` and a dependent array `dependent_values` search domain for value closest to 
    `desired_x` and return the value of `dependent_values` at that location if it is defined. Otherwise, use local 
    slopes of `desired_x` and `dependent_values` to interpolate a value of `dependent_values` at `desired_x`.

    Based on `numpy`'s `interp` function.

    Parameters
    ----------
    desired_x : float
        Location where `dependent_variables` is desired.
    x_domain : np.ndarray[float]
        Domain to search for the correct location.
    dependent_values : np.ndarray[complex]
        Dependent values that are to be returned after search and interpolation.
    provided_j : int
        Give a j index from a previous interpolation to improve performance.

    Returns
    -------
    result : complex
        Desired value of `dependent_values`.
    j_out : int
        The index that was found during c_binary_search_with_guess which can be used by other interpolation functions 
        to improve performance.

    """
    cdef Py_ssize_t len_x
    len_x = len(x_domain)

    return interp_complexj_ptr(desired_x, &x_domain[0], &dependent_values[0], len_x, provided_j)


cdef void interp_array_ptr(
        const double* desired_x_array,
        const double* x_domain,
        const double* dependent_values,
        double* desired_dependent_array,
        Py_ssize_t len_x,
        Py_ssize_t desired_len
        ) noexcept nogil:

    # Array variables
    cdef Py_ssize_t index

    # Binary Search with Guess
    cdef double result

    # Since most of the use cases for this function are subsampling an array with another array, we can improve our
    #  guess by increasing it alongside the index variable. There are problems with this:
    #    1 - If the desired array is randomly distributed, rather than increasing, this will be slow.
    #    2 - The actual x domain is likely not linear. So the linear increase we are performing with this guess variable
    #        is not correct.
    cdef double x_slope
    cdef Py_ssize_t guess, j
    x_slope = <double>len_x / <double>desired_len
    if x_slope < 1.:
        x_slope = 1.

    cdef double desired_x
    for index in range(desired_len):
        desired_x = desired_x_array[index]

        # Perform binary search with guess
        guess = <Py_ssize_t>floor(x_slope * <double>index)
        j = c_binary_search_with_guess(desired_x, x_domain, len_x, guess)

        # Run interpolation
        result = interp_ptr(desired_x, x_domain, dependent_values, len_x, provided_j=j)

        # Store result
        desired_dependent_array[index] = result


cpdef void interp_array(
        const double[::1] desired_x_array,
        const double[::1] x_domain,
        const double[::1] dependent_values,
        double[::1] desired_dependent_array
        ) noexcept nogil:

    # Array variables
    cdef Py_ssize_t len_x, desired_len
    len_x       = len(x_domain)
    desired_len = len(desired_x_array)

    interp_array_ptr(&desired_x_array[0], &x_domain[0], &dependent_values[0], &desired_dependent_array[0],
                     len_x, desired_len)


cdef void interp_complex_array_ptr(
        const double* desired_x_array,
        const double* x_domain,
        const  double complex* dependent_values,
        double complex* desired_dependent_array,
        Py_ssize_t len_x,
        Py_ssize_t desired_len
        ) noexcept nogil:

    # Array variables
    cdef Py_ssize_t index

    # Since most of the use cases for this function are subsampling an array with another array, we can improve our
    #  guess by increasing it alongside the index variable. There are problems with this:
    #    1 - If the desired array is randomly distributed, rather than increasing, this will be slow.
    #    2 - The actual x domain is likely not linear. So the linear increase we are performing with this guess variable
    #        is not correct.
    cdef double x_slope
    cdef Py_ssize_t guess, j
    x_slope = <double>len_x / <double>desired_len
    if x_slope < 1.:
        x_slope = 1.

    cdef double desired_x
    cdef double complex result
    for index in range(desired_len):
        desired_x = desired_x_array[index]

        # Perform binary search with guess
        guess = <Py_ssize_t>floor(x_slope * <double>index)
        j = c_binary_search_with_guess(desired_x, x_domain, len_x, guess)

        # Run interpolation
        result = interp_complex_ptr(desired_x, x_domain, dependent_values, len_x, provided_j=j)

        # Store result
        desired_dependent_array[index] = result


cpdef void interp_complex_array(
        const double[::1] desired_x_array,
        const double[::1] x_domain,
        const double complex[::1] dependent_values,
        double complex[::1] desired_dependent_array
        ) noexcept nogil:

    cdef Py_ssize_t len_x, desired_len
    len_x       = len(x_domain)
    desired_len = len(desired_x_array)

    interp_complex_array_ptr(&desired_x_array[0], &x_domain[0], &dependent_values[0], &desired_dependent_array[0],
                             len_x, desired_len)


def binary_search_with_guess(double key, double[::1] array, ssize_t length, ssize_t guess):

    cdef double* array_ptr = &array[0]
    return c_binary_search_with_guess(key, array_ptr, length, guess)


import numpy as np
from numba import njit

from CyRK import nb2cy, cy2nb, nbsolve_ivp, pysolve_ivp


@njit
def diffeq_cy(dy, t, y):
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]


@njit
def diffeq_cy_args(dy, t, y, a, b):
    dy[0] = (1. - a * y[1]) * y[0]
    dy[1] = (b * y[0] - 1.) * y[1]


@njit
def diffeq_scipy(t, y):
    dy = np.zeros_like(y)
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]
    return dy


@njit
def diffeq_scipy_args(t, y, a, b):
    dy = np.zeros_like(y)
    dy[0] = (1. - a * y[1]) * y[0]
    dy[1] = (b * y[0] - 1.) * y[1]
    return dy


initial_conds = np.asarray((20., 20.), dtype=np.float64)
time_span = (0., 20.)
rtol = 1.0e-7
atol = 1.0e-8
t_eval = np.linspace(time_span[0], time_span[1], 10)


def test_nb2cy_noargs():
    """ Test converting a scipy diffeq into a cyrk one with no additional arguments. """

    # Accuracy check tolerances
    check_rtol = 0.1

    # First calculate the result of an integration using nbrk.
    result_nb = \
        nbsolve_ivp(diffeq_scipy, time_span, initial_conds, t_eval=t_eval, rtol=rtol, atol=atol)
    assert result_nb.success

    # Perform a cyrk integration using the diffeq that was written for cyrk
    result_cy = \
        pysolve_ivp(diffeq_cy, time_span, initial_conds,
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)

    assert result_cy.success

    # Perform the function conversion
    diffeq_cy_converted = nb2cy(diffeq_scipy)

    # Use this function to recalculate using cyrk
    result_cy_conv = \
        pysolve_ivp(diffeq_cy_converted, time_span, initial_conds,
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy_conv.success

    # Check that the results match
    # # Converted vs. hardcoded
    p_diff_1_cy = 2. * np.abs((result_cy_conv.y[0] - result_cy.y[0]) / (result_cy_conv.y[0] + result_cy.y[0]))
    p_diff_2_cy = 2. * np.abs((result_cy_conv.y[1] - result_cy.y[1]) / (result_cy_conv.y[1] + result_cy.y[1]))
    assert np.all(p_diff_1_cy < check_rtol)
    assert np.all(p_diff_2_cy < check_rtol)

    # # Converted vs. nbrk
    p_diff_1_nb = 2. * np.abs((result_cy_conv.y[0] - result_nb.y[0]) / (result_cy_conv.y[0] + result_nb.y[0]))
    p_diff_2_nb = 2. * np.abs((result_cy_conv.y[1] - result_nb.y[1]) / (result_cy_conv.y[1] + result_nb.y[1]))
    assert np.all(p_diff_1_nb < check_rtol)
    assert np.all(p_diff_2_nb < check_rtol)


def test_nb2cy_args():
    """ Test converting a scipy diffeq into a cyrk one with additional arguments. """

    # Accuracy check tolerances
    check_rtol = 0.1

    # First calculate the result of an integration using nbrk.
    result_nb = \
        nbsolve_ivp(diffeq_scipy_args, time_span, initial_conds, t_eval=t_eval, args=(0.01, 0.02), rtol=rtol, atol=atol)
    assert result_nb.success

    # Perform a cyrk integration using the diffeq that was written for cyrk
    result_cy = \
        pysolve_ivp(diffeq_cy_args, time_span, initial_conds,
                    args=(0.01, 0.02),
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy.success

    # Perform the function conversion
    diffeq_cy_converted_args = nb2cy(diffeq_scipy_args)

    # Use this function to recalculate using cyrk
    result_cy_conv = \
        pysolve_ivp(diffeq_cy_converted_args, time_span, initial_conds,
                    args=(0.01, 0.02),
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy_conv.success

    # Check that the results match
    # # Converted vs. hardcoded
    p_diff_1_cy = 2. * np.abs((result_cy_conv.y[0] - result_cy.y[0]) / (result_cy_conv.y[0] + result_cy.y[0]))
    p_diff_2_cy = 2. * np.abs((result_cy_conv.y[1] - result_cy.y[1]) / (result_cy_conv.y[1] + result_cy.y[1]))
    assert np.all(p_diff_1_cy < check_rtol)
    assert np.all(p_diff_2_cy < check_rtol)

    # # Converted vs. nbrk
    p_diff_1_nb = 2. * np.abs((result_cy_conv.y[0] - result_nb.y[0]) / (result_cy_conv.y[0] + result_nb.y[0]))
    p_diff_2_nb = 2. * np.abs((result_cy_conv.y[1] - result_nb.y[1]) / (result_cy_conv.y[1] + result_nb.y[1]))
    assert np.all(p_diff_1_nb < check_rtol)
    assert np.all(p_diff_2_nb < check_rtol)


def test_cy2nb_noargs():
    """ Test converting a cyrk diffeq into a numba/scipy one with no additional arguments. """

    # Accuracy check tolerances
    check_rtol = 0.1

    # First calculate the result of an integration using nbrk.
    result_nb = \
        nbsolve_ivp(diffeq_scipy, time_span, initial_conds, t_eval=t_eval, rtol=rtol, atol=atol)
    assert result_nb.success

    # Perform a cyrk integration using the diffeq that was written for cyrk
    result_cy = \
        pysolve_ivp(diffeq_cy, time_span, initial_conds,
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy.success

    # Perform the function conversion
    diffeq_nb_converted = cy2nb(diffeq_cy)

    # Use this function to recalculate using cyrk
    result_nb_conv = \
        nbsolve_ivp(diffeq_nb_converted, time_span, initial_conds, t_eval=t_eval, rtol=rtol, atol=atol)
    assert result_nb.success

    # Check that the results match
    # # Converted vs. hardcoded
    p_diff_1_cy = 2. * np.abs((result_nb_conv.y[0] - result_cy.y[0]) / (result_nb_conv.y[0] + result_cy.y[0]))
    p_diff_2_cy = 2. * np.abs((result_nb_conv.y[1] - result_cy.y[1]) / (result_nb_conv.y[1] + result_cy.y[1]))
    assert np.all(p_diff_1_cy < check_rtol)
    assert np.all(p_diff_2_cy < check_rtol)

    # # Converted vs. nbrk
    p_diff_1_nb = 2. * np.abs((result_nb_conv.y[0] - result_nb.y[0]) / (result_nb_conv.y[0] + result_nb.y[0]))
    p_diff_2_nb = 2. * np.abs((result_nb_conv.y[1] - result_nb.y[1]) / (result_nb_conv.y[1] + result_nb.y[1]))
    assert np.all(p_diff_1_nb < check_rtol)
    assert np.all(p_diff_2_nb < check_rtol)


def test_cy2nb_args():
    """ Test converting a cyrk diffeq into a numba/scipy one with additional arguments. """

    # Accuracy check tolerances
    check_rtol = 0.1

    # First calculate the result of an integration using nbrk.
    result_nb = \
        nbsolve_ivp(diffeq_scipy_args, time_span, initial_conds, t_eval=t_eval, args=(0.01, 0.02), rtol=rtol, atol=atol)
    assert result_nb.success

    # Perform a cyrk integration using the diffeq that was written for cyrk
    result_cy = \
        pysolve_ivp(diffeq_cy_args, time_span, initial_conds,
                    args=(0.01, 0.02),
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy.success

    # Perform the function conversion
    diffeq_nb_converted_args = cy2nb(diffeq_cy_args)

    # Use this function to recalculate using cyrk
    nb_result_conv = \
        nbsolve_ivp(diffeq_nb_converted_args, time_span, initial_conds, t_eval=t_eval, args=(0.01, 0.02), rtol=rtol, atol=atol)
    assert nb_result_conv.success

    # Check that the results match
    # # Converted vs. hardcoded
    p_diff_1_cy = 2. * np.abs((nb_result_conv.y[0] - result_cy.y[0]) / (nb_result_conv.y[0] + result_cy.y[0]))
    p_diff_2_cy = 2. * np.abs((nb_result_conv.y[1] - result_cy.y[1]) / (nb_result_conv.y[1] + result_cy.y[1]))
    assert np.all(p_diff_1_cy < check_rtol)
    assert np.all(p_diff_2_cy < check_rtol)

    # # Converted vs. nbrk
    p_diff_1_nb = 2. * np.abs((nb_result_conv.y[0] - result_nb.y[0]) / (nb_result_conv.y[0] + result_nb.y[0]))
    p_diff_2_nb = 2. * np.abs((nb_result_conv.y[1] - result_nb.y[1]) / (nb_result_conv.y[1] + result_nb.y[1]))
    assert np.all(p_diff_1_nb < check_rtol)
    assert np.all(p_diff_2_nb < check_rtol)


def test_cy2nb_cache_njit():
    """ Test converting a cyrk diffeq into a numba/scipy with njit cacheing on. """

    # Accuracy check tolerances
    check_rtol = 0.1

    # First calculate the result of an integration using nbrk.
    result_nb = \
        nbsolve_ivp(diffeq_scipy, time_span, initial_conds, t_eval=t_eval, rtol=rtol, atol=atol)
    assert result_nb.success

    # Perform a cyrk integration using the diffeq that was written for cyrk
    result_cy = \
        pysolve_ivp(diffeq_cy, time_span, initial_conds,
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy.success

    # Perform the function conversion
    diffeq_nb_converted = cy2nb(diffeq_cy)

    # Use this function to recalculate using cyrk
    nb_result_conv = \
        nbsolve_ivp(diffeq_nb_converted, time_span, initial_conds, t_eval=t_eval, rtol=rtol, atol=atol)
    assert nb_result_conv.success

    # Check that the results match
    # # Converted vs. hardcoded
    p_diff_1_cy = 2. * np.abs((nb_result_conv.y[0] - result_cy.y[0]) / (nb_result_conv.y[0] + result_cy.y[0]))
    p_diff_2_cy = 2. * np.abs((nb_result_conv.y[1] - result_cy.y[1]) / (nb_result_conv.y[1] + result_cy.y[1]))
    assert np.all(p_diff_1_cy < check_rtol)
    assert np.all(p_diff_2_cy < check_rtol)

    # # Converted vs. nbrk
    p_diff_1_nb = 2. * np.abs((nb_result_conv.y[0] - result_nb.y[0]) / (nb_result_conv.y[0] + result_nb.y[0]))
    p_diff_2_nb = 2. * np.abs((nb_result_conv.y[1] - result_nb.y[1]) / (nb_result_conv.y[1] + result_nb.y[1]))
    assert np.all(p_diff_1_nb < check_rtol)
    assert np.all(p_diff_2_nb < check_rtol)


def test_nb2cy_cache_njit():
    """ Test converting a scipy diffeq into a cyrk with njit cacheing on. """

    # Accuracy check tolerances
    check_rtol = 0.1

    # First calculate the result of an integration using nbrk.
    result_nb = \
        nbsolve_ivp(diffeq_scipy, time_span, initial_conds, t_eval=t_eval, rtol=rtol, atol=atol)
    assert result_nb.success

    # Perform a cyrk integration using the diffeq that was written for cyrk
    result_cy = \
        pysolve_ivp(diffeq_cy, time_span, initial_conds,
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy.success

    # Perform the function conversion
    diffeq_cy_converted = nb2cy(diffeq_scipy)

    # Use this function to recalculate using cyrk
    result_cy_conv = \
        pysolve_ivp(diffeq_cy_converted, time_span, initial_conds,
                    rtol=rtol, atol=atol, t_eval=t_eval,
                    pass_dy_as_arg=True)
    assert result_cy_conv.success

    # Check that the results match
    # # Converted vs. hardcoded
    p_diff_1_cy = 2. * np.abs((result_cy_conv.y[0] - result_cy.y[0]) / (result_cy_conv.y[0] + result_cy.y[0]))
    p_diff_2_cy = 2. * np.abs((result_cy_conv.y[1] - result_cy.y[1]) / (result_cy_conv.y[1] + result_cy.y[1]))
    assert np.all(p_diff_1_cy < check_rtol)
    assert np.all(p_diff_2_cy < check_rtol)

    # # Converted vs. nbrk
    p_diff_1_nb = 2. * np.abs((result_cy_conv.y[0] - result_nb.y[0]) / (result_cy_conv.y[0] + result_nb.y[0]))
    p_diff_2_nb = 2. * np.abs((result_cy_conv.y[1] - result_nb.y[1]) / (result_cy_conv.y[1] + result_nb.y[1]))
    assert np.all(p_diff_1_nb < check_rtol)
    assert np.all(p_diff_2_nb < check_rtol)



import numpy as np
from numba import njit

from CyRK import nbsolve_ivp


@njit
def diffeq_extra_outputs(t, y):
    extra_0 = (1. - 0.01 * y[1])
    extra_1 = (0.02 * y[0] - 1.)
    dy_0 = extra_0 * y[0]
    dy_1 = extra_1 * y[1]
    return np.asarray([dy_0, dy_1, extra_0, extra_1], dtype=y.dtype)


initial_conds = np.asarray((20., 20.), dtype=np.complex128)
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-7
atol = 1.0e-8


def test_extra_output_integration():
    """Check that the numba solver is able to run and capture additional outputs """

    nbrk_result = \
        nbsolve_ivp(diffeq_extra_outputs, time_span, initial_conds, capture_extra=True)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    assert nbrk_result.y.dtype == np.complex128
    assert nbrk_result.t.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y.shape[0] == 4
    assert nbrk_result.y[0].size == nbrk_result.y[1].size
    assert nbrk_result.y[0].size == nbrk_result.y[2].size
    assert nbrk_result.y[0].size == nbrk_result.y[3].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

def test_extra_output_integration_teval_no_extra_interpolation():
    """Check that the numba solver is able to run and capture additional outputs.
    Reduced t_eval used but no interpolation used on the extra parameters.
    """

    t_eval = np.linspace(time_span[0], time_span[1], 5)

    nbrk_result = \
        nbsolve_ivp(
            diffeq_extra_outputs, time_span, initial_conds, t_eval=t_eval,
            capture_extra=True, interpolate_extra=False)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    assert nbrk_result.y.dtype == np.complex128
    assert nbrk_result.t.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y.shape[0] == 4
    assert nbrk_result.y[0].size == nbrk_result.y[1].size
    assert nbrk_result.y[0].size == nbrk_result.y[2].size
    assert nbrk_result.y[0].size == nbrk_result.y[3].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str


def test_extra_output_integration_teval_with_extra_interpolation():
    """Check that the numba solver is able to run and capture additional outputs
    Reduced t_eval used with interpolation used on the extra parameters.
    """

    t_eval = np.linspace(time_span[0], time_span[1], 5)

    nbrk_result = \
        nbsolve_ivp(
                diffeq_extra_outputs, time_span, initial_conds, t_eval=t_eval,
                capture_extra=True, interpolate_extra=True)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    assert nbrk_result.y.dtype == np.complex128
    assert nbrk_result.t.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y.shape[0] == 4
    assert nbrk_result.y[0].size == nbrk_result.y[1].size
    assert nbrk_result.y[0].size == nbrk_result.y[2].size
    assert nbrk_result.y[0].size == nbrk_result.y[3].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str



import pytest
import numpy as np
from scipy.integrate import solve_ivp
from numba import njit

from CyRK import nbsolve_ivp


@njit
def diffeq_args(t, y, a, b):
    dy = np.zeros_like(y)
    dy[0] = (1. - a * y[1]) * y[0]
    dy[1] = (b * y[0] - 1.) * y[1]
    return dy


@njit
def diffeq(t, y):
    dy = np.zeros_like(y)
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]
    return dy


initial_conds = np.asarray((20., 20.), dtype=np.float64)
initial_conds_complex = np.asarray((20. + 0.01j, 20. - 0.01j), dtype=np.complex128)
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-7
atol = 1.0e-8

rtols = np.asarray((1.0e-7, 1.0e-8), dtype=np.float64, order='C')
atols = np.asarray((1.0e-8, 1.0e-9), dtype=np.float64, order='C')


def test_nbrk_test():
    """Check that the builtin test function for the nbrk integrator is working"""

    from CyRK import test_nbrk
    test_nbrk()

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
@pytest.mark.parametrize('use_rtol_array', (True, False))
@pytest.mark.parametrize('use_atol_array', (True, False))
def test_basic_integration(use_atol_array, use_rtol_array, rk_method, complex_valued):
    """Check that the numba solver is able to run with its default arguments"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    tol_dict = dict()
    if use_atol_array:
        tol_dict['atols'] = atols
    else:
        tol_dict['atol'] = atol
    if use_rtol_array:
        tol_dict['rtols'] = rtols
    else:
        tol_dict['rtol'] = rtol

    nbrk_result = \
        nbsolve_ivp(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, **tol_dict)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_different_tols(rk_method, complex_valued):
    """Check that the numba solver is able to run with different tolerances"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    nbrk_result = \
        nbsolve_ivp(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, rtol=1.0e-10, atol=1.0e-12)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_max_step(rk_method, complex_valued):
    """Check that the numba solver is able to run with different maximum step size"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    nbrk_result = \
        nbsolve_ivp(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, max_step=time_span[1] / 2.)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_first_step(rk_method, complex_valued):
    """Check that the numba solver is able to run with a user provided first step size"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    nbrk_result = \
        nbsolve_ivp(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, first_step=0.01)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_large_end_value(rk_method, complex_valued):
    """Check that the numba solver is able to run using a larger ending time value """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    nbrk_result = \
        nbsolve_ivp(diffeq, time_span_large, initial_conds_to_use, rk_method=rk_method)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_teval(rk_method, complex_valued):
    """Check that the numba solver is able to run using a user provided t_eval """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    t_eval = np.linspace(time_span[0], time_span[1], 10)

    nbrk_result = \
        nbsolve_ivp(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, t_eval=t_eval)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_args(rk_method, complex_valued):
    """Check that the numba solver is able to run with user provided additional diffeq arguments """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    nbrk_result = \
        nbsolve_ivp(diffeq_args, time_span, initial_conds_to_use, rk_method=rk_method, args=(0.01, 0.02))

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_accuracy(rk_method):
    """Check that the numba solver is able to reproduce a known functions integral with reasonable accuracy """

    # TODO: This is only checking one equation. Add other types of diffeqs to provide better coverage.

    # Differential Equation
    @njit
    def diffeq_accuracy(t, y):
        dy = np.empty_like(y)
        dy[0] = np.sin(t) - y[1]  # dydt = sin(t) - x(t)
        dy[1] = np.cos(t) + y[0]  # dxdt = cos(t) + y(t)
        return dy

    @njit
    def correct_answer(t, c1_, c2_):
        y = np.empty((2, t.size), dtype=np.float64)
        y[0] = -c1_ * np.sin(t) + c2_ * np.cos(t) - (np.cos(t) / 2)  # -c1 * sin(t) + c2 * cos(t) - cos(t) / 2
        # At t=0; y = c2 - 1/2
        y[1] = c2_ * np.sin(t) + c1_ * np.cos(t) + (np.sin(t) / 2)   # c2 * sin(t) + c1 * cos(t) + sin(t) / 2
        # At t=0; x = c1
        return y

    # Initial Conditions
    # y=0 --> c2 = + 1/2
    c2 = 0.5
    # x=1 --> c1 = + 1
    c1 = 1.0
    y0 = np.asarray((0., 1.), dtype=np.float64)
    time_span_ = (0., 10.)

    # CyRK.nbsolve_ivp
    nbrk_result = \
        nbsolve_ivp(diffeq_accuracy, time_span_, y0, rk_method=rk_method, rtol=rtol, atol=atol)
    real_answer = correct_answer(nbrk_result.t, c1, c2)

    if rk_method == 0:
        assert np.allclose(nbrk_result.y, real_answer, rtol=1.0e-3, atol=1.0e-6)
    elif rk_method == 1:
        assert np.allclose(nbrk_result.y, real_answer, rtol=1.0e-4, atol=1.0e-7)
    else:
        assert np.allclose(nbrk_result.y, real_answer, rtol=1.0e-5, atol=1.0e-8)

    # Check the accuracy of the results
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # ax.plot(nbrk_result.t, nbrk_result.y[0], 'r', label='CyRK')
    # ax.plot(nbrk_result.t, nbrk_result.y[1], 'r:')
    # ax.plot(nbrk_result.t, real_answer[0], 'b', label='Analytic')
    # ax.plot(nbrk_result.t, real_answer[1], 'b:')
    # plt.show()

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_max_num_steps(rk_method, complex_valued):
    """Check that the numba solver is able to utilize the max_num_steps argument """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    # First test a number of max steps which is fine.
    nbrk_result = \
        nbsolve_ivp(diffeq, time_span_large, initial_conds_to_use, rk_method=rk_method, max_num_steps=1000000)

    # Check that the ndarrays make sense
    assert type(nbrk_result.t) == np.ndarray
    assert nbrk_result.t.dtype == np.float64
    if complex_valued:
        assert nbrk_result.y.dtype == np.complex128
    else:
        assert nbrk_result.y.dtype == np.float64
    assert nbrk_result.t.size > 1
    assert nbrk_result.t.size == nbrk_result.y[0].size
    assert len(nbrk_result.y.shape) == 2
    assert nbrk_result.y[0].size == nbrk_result.y[1].size

    # Check that the other output makes sense
    assert type(nbrk_result.success) == bool
    assert nbrk_result.success
    assert type(nbrk_result.message) == str

    # Now test an insufficient number of steps
    nbrk_result = \
        nbsolve_ivp(diffeq, time_span_large, initial_conds_to_use, rk_method=rk_method, max_num_steps=4)

    # Check that the ndarrays make sense
    assert not nbrk_result.success
    assert nbrk_result.message == "Maximum number of steps (set by user) exceeded during integration."



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

import numpy as np
cimport numpy as cnp
cnp.import_array()

# =====================================================================================================================
# Import CySolverResult (container for integration results)
# =====================================================================================================================
cdef class WrapCySolverResult:

    cdef void set_cyresult_pointer(self, shared_ptr[CySolverResult] cyresult_shptr):

        # Store c++ based result and pull out key information
        self.cyresult_shptr = cyresult_shptr
        self.cyresult_ptr   = cyresult_shptr.get()

        # Convert solution to pointers and views
        if self.cyresult_ptr.size > 0:
            self.time_ptr  = &self.cyresult_ptr.time_domain_vec[0]
            self.y_ptr     = &self.cyresult_ptr.solution[0]
            self.time_view = <double[:self.size]>self.time_ptr
            self.y_view    = <double[:self.size * self.num_dy]>self.y_ptr
    
    def call(self, double t):
        """ Call the dense output interpolater and return y """

        if not self.cyresult_ptr.capture_dense_output:
            raise AttributeError("Can not call WrapCySolverResult when dense_output set to False.")

        y_interp_array = np.empty(self.cyresult_ptr.num_dy, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr      = &y_interp_view[0]

        self.cyresult_ptr.call(t, y_interp_ptr)
        return y_interp_array
    
    def call_vectorize(self, double[::1] t_view):
        """ Call the dense output interpolater and return y """

        if not self.cyresult_ptr.capture_dense_output:
            raise AttributeError("Can not call WrapCySolverResult when dense_output set to False.")

        cdef size_t len_t = len(t_view)

        y_interp_array = np.empty(self.cyresult_ptr.num_dy * len_t, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr = &y_interp_view[0]
        cdef double* t_array_ptr  = &t_view[0]

        self.cyresult_ptr.call_vectorize(t_array_ptr, len_t, y_interp_ptr)
        return y_interp_array.reshape(len_t, self.cyresult_ptr.num_dy).T

    @property
    def success(self):
        return self.cyresult_ptr.success
        
    @property
    def message(self):
        return str(self.cyresult_ptr.message_ptr, 'UTF-8')
    
    @property
    def t(self):
        return np.asarray(self.time_view, dtype=np.float64, order='C')
    
    @property
    def y(self):
        return np.asarray(self.y_view, dtype=np.float64, order='C').reshape((self.size, self.num_dy)).T
    
    @property
    def size(self):
        return self.cyresult_ptr.size
    
    @property
    def num_y(self):
        return self.cyresult_ptr.num_y
    
    @property
    def num_dy(self):
        return self.cyresult_ptr.num_dy
    
    @property
    def error_code(self):
        return self.cyresult_ptr.error_code
    
    def __call__(self, t):

        if type(t) == cnp.ndarray:
            return self.call_vectorize(t)
        else:
            return self.call(t).reshape(self.cyresult_ptr.num_dy, 1)


# =====================================================================================================================
# Create Wrapped cysolve_ivp (has various defaults)
# =====================================================================================================================
cdef void cysolve_ivp_noreturn(
            shared_ptr[CySolverResult] solution_sptr,
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            void* args_ptr = NULL,
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            PreEvalFunc pre_eval_func = NULL,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept nogil:
    baseline_cysolve_ivp_noreturn(
        solution_sptr,
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        pre_eval_func,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

cdef CySolveOutput cysolve_ivp(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            void* args_ptr = NULL,
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            PreEvalFunc pre_eval_func = NULL,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept nogil:
    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        pre_eval_func,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

    return result

cdef CySolveOutput cysolve_ivp_gil(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            void* args_ptr = NULL,
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            PreEvalFunc pre_eval_func = NULL,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept:
    
    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        pre_eval_func,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

    return result



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

cdef public api void call_diffeq_from_cython(object py_instance, DiffeqMethod diffeq):
    """Callback function used by the C++ model.
    The "public api" prefix tells Cython to produce header files "pysolver_api.h" which can be included in
    C++ files.
    """

    # Call the python diffeq.
    diffeq(py_instance)



""" Lorenz Differential Equations """

import numpy as np
from numba import njit

from CyRK import nb2cy

lorenz_y0 = np.asarray((1., 0., 0.), dtype=np.float64)
lorenz_args = (10., 28.0, 8. / 3.)
lorenz_time_span_1 = (0., 10.)
lorenz_time_span_2 = (0., 100.)


@njit(cache=True)
def lorenz_nb(t, y, a, b, c):

    y0 = y[0]
    y1 = y[1]
    y2 = y[2]
    dy = np.empty_like(y)
    dy[0] = a * (y1 - y0)
    dy[1] = y0 * (b - y2) - y1
    dy[2] = y0 * y1 - c * y2
    return dy


lorenz_cy = nb2cy(lorenz_nb, use_njit=True, cache_njit=True)

@njit(cache=True)
def lorenz_nb_extra(t, y, a, b, c):

    y0 = y[0]
    y1 = y[1]
    y2 = y[2]

    e_1 = a
    e_2 = (b - y2)
    e_3 = c * y2

    dy_0 = e_1 * (y1 - y0)
    dy_1 = y0 * e_2 - y1
    dy_2 = y0 * y1 - e_3
    return np.asarray([dy_0, dy_1, dy_2, e_1, e_2, e_3], dtype=np.float64)


lorenz_cy_extra = nb2cy(lorenz_nb_extra, use_njit=True, cache_njit=True)



from CyRK.utils.vector cimport vector
from CyRK.cy.cysolver_api cimport CySolveOutput, CySolverResult

cdef void interpolate_from_solution_list(
    double* y_result_ptr,
    int num_y,
    vector[CySolveOutput] solution_list_vec,
    int num_solutions,
    double* x_domain_ptr,
    size_t x_domain_size,
    vector[double] x_breakpoints_vec) noexcept nogil



""" Commands to build the cython extensions of CyRK (a hack to work with pyproject.toml) """
import os
import platform
import math
import json
import sys
from setuptools.extension import Extension
from setuptools.command.build_py import build_py as _build_py
from setuptools.command.build_ext import build_ext as _build_ext

import numpy as np

num_procs = os.cpu_count()
num_threads = max(1, num_procs - 1)

install_platform = platform.system()

if install_platform.lower() == 'windows':
    extra_compile_args = ['/openmp']
    extra_link_args = []
elif install_platform.lower() == 'darwin':
    # OpenMP is installed via llvm. See https://stackoverflow.com/questions/60005176/how-to-deal-with-clang-error-unsupported-option-fopenmp-on-travis
    extra_compile_args = ['-O3', '-fopenmp']
    extra_link_args = ['-lomp']
else:
    extra_compile_args = ['-fopenmp', '-O3']
    extra_link_args = ['-fopenmp', '-O3']
macro_list = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]

# Load CyRK's cython extensions
absolute_path = os.path.dirname(__file__)
cython_ext_path = os.path.join(absolute_path, 'cython_extensions.json')
with open(cython_ext_path, 'r') as cython_ext_file:
    cython_ext_dict = json.load(cython_ext_file)

cython_extensions = list()
for cython_ext, ext_data in cython_ext_dict.items():

    if ext_data['is_cpp']:
        if install_platform.lower() == 'windows':
            specific_compile_args = extra_compile_args + ext_data['compile_args'] + ["/std:c++20"]
        else:
            specific_compile_args = extra_compile_args + ext_data['compile_args'] + ["-std=c++20"]
    else:
        specific_compile_args = extra_compile_args + ext_data['compile_args']

    cython_extensions.append(
        Extension(
            name=ext_data['name'],
            sources=[os.path.join(*tuple(source_path)) for source_path in ext_data['sources']],
            # Always add numpy to any includes; also add sys.path so we can capture python.h
            include_dirs=[os.path.join(*tuple(dir_path)) for dir_path in ext_data['include_dirs']] + [np.get_include()] + sys.path,
            extra_compile_args=specific_compile_args,
            define_macros=macro_list,
            extra_link_args=ext_data['link_args'] + extra_link_args,
            )
        )

class build_ext(_build_ext):

    def run(self):
        # Compile in parallel
        self.parallel = num_threads
        return super().run()


class build_cyrk(_build_py):

    def run(self):
        # Run build_ext to compile cythonized c/cpp files.
        self.run_command("build_ext")
        return super().run()

    def initialize_options(self):
        super().initialize_options()
        from Cython.Build import cythonize
        print('!-- Cythonizing CyRK')
        if self.distribution.ext_modules == None:
            self.distribution.ext_modules = []

        # Add cython extensions to ext_modules list
        for extension in cython_extensions:
            self.distribution.ext_modules.append(
                    extension
                    )

        # Cythonize ext_modules
        self.distribution.ext_modules = cythonize(
                self.distribution.ext_modules,
                compiler_directives={'language_level': "3"},
                include_path=['.', np.get_include()],
                nthreads=num_threads,
                )
        print('!-- Finished Cythonizing CyRK')



import numpy as np
import pytest
from numba import njit

from CyRK import pysolve_ivp, WrapCySolverResult


# To reduce number of tests, only test RK23 once since RK45 should capture all its functionality
SKIP_SOME_RK23_TESTS = True
RK23_TESTED = False

def diffeq(dy, t, y):
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]


def diffeq_stiff(t, y):
    # TODO: Replace with function that is actually stiff
    dy = np.empty(y.size, dtype=np.float64)
    dy[0] = (1. - 0.01 * y[1]) * np.exp(y[0])
    dy[1] = (0.02 * y[0] - 1.) * np.exp(y[1])
    return dy

def diffeq_args(dy, t, y, a, b):
    dy[0] = (1. - a * y[1]) * y[0]
    dy[1] = (b * y[0] - 1.) * y[1]

def diffeq_scipy_style(t, y):
    dy = np.empty(y.size, dtype=np.float64)
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]
    return dy

def diffeq_scipy_style_args(t, y, a, b):
    dy = np.empty(y.size, dtype=np.float64)
    dy[0] = (1. - a * y[1]) * y[0]
    dy[1] = (b * y[0] - 1.) * y[1]
    return dy

def diffeq_extra(dy, t, y):
    extra_1 = (1. - 0.01 * y[1])
    extra_2 = (0.02 * y[0] - 1.)
    dy[0] = extra_1 * y[0]
    dy[1] = extra_2 * y[1]
    dy[2] = extra_1
    dy[3] = extra_2

def diffeq_args_extra(dy, t, y, a, b):
    extra_1 = (1. - 0.01 * y[1])
    extra_2 = (0.02 * y[0] - 1.)
    dy[0] = extra_1 * y[0]
    dy[1] = extra_2 * y[1]
    dy[2] = extra_1
    dy[3] = extra_2

def diffeq_scipy_style_extra(t, y):
    dy = np.empty(y.size + 2, dtype=np.float64)
    extra_1 = (1. - 0.01 * y[1])
    extra_2 = (0.02 * y[0] - 1.)
    dy[0] = extra_1 * y[0]
    dy[1] = extra_2 * y[1]
    dy[2] = extra_1
    dy[3] = extra_2
    return dy

def diffeq_scipy_style_args_extra(t, y, a, b):
    dy = np.empty(y.size + 2, dtype=np.float64)
    extra_1 = (1. - 0.01 * y[1])
    extra_2 = (0.02 * y[0] - 1.)
    dy[0] = extra_1 * y[0]
    dy[1] = extra_2 * y[1]
    dy[2] = extra_1
    dy[3] = extra_2
    return dy

args = (0.01, 0.02)

initial_conds = np.asarray((20., 20.), dtype=np.float64, order='C')
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-4
atol = 1.0e-7

rtols = np.asarray((1.0e-4, 1.0e-5), dtype=np.float64, order='C')
atols = np.asarray((1.0e-6, 1.0e-7), dtype=np.float64, order='C')


def test_pysolve_ivp_test():
    """Check that the builtin test function for the PySolver integrator is working"""

    from CyRK import test_pysolver
    test_pysolver()

# njit is slow during testing so only do it once for each diffeq
njit_rk23_tested = False
njit_rk45_tested = False
njit_DOP853_tested = False

@pytest.mark.filterwarnings("error")  # Some exceptions get propagated via cython as warnings; we want to make sure the lead to crashes.
@pytest.mark.parametrize('capture_extra', (True, False))
@pytest.mark.parametrize('max_step', (1.0, 100_000.0))
@pytest.mark.parametrize('first_step', (0.0, 0.00001))
@pytest.mark.parametrize('integration_method', ("RK23", "RK45", "DOP853"))
@pytest.mark.parametrize('use_different_tols', (True, False))
@pytest.mark.parametrize('use_rtol_array', (True, False))
@pytest.mark.parametrize('use_atol_array', (True, False))
@pytest.mark.parametrize('use_large_timespan', (True, False))
@pytest.mark.parametrize('use_njit_always', (False,))
@pytest.mark.parametrize('use_args', (True, False))
@pytest.mark.parametrize('use_scipy_style', (True, False))
def test_pysolve_ivp(use_scipy_style, use_args, use_njit_always,
                     use_large_timespan, use_atol_array, use_rtol_array, use_different_tols, integration_method,
                     first_step, max_step, capture_extra):
    """Check that the pysolve_ivp function is able to run with various changes to its arguments. """
    global RK23_TESTED
    global njit_rk23_tested
    global njit_rk45_tested
    global njit_DOP853_tested

    # To reduce number of tests, only test RK23 once. 
    if RK23_TESTED and SKIP_SOME_RK23_TESTS and (integration_method=="RK23"):
        pytest.skip("Skipping Some RK23 Tests (just to reduce number of tests).")
    else:
        RK23_TESTED = True

    if use_args:
        if use_scipy_style:
            if capture_extra:
                diffeq_to_use = diffeq_scipy_style_args_extra
            else:
                diffeq_to_use = diffeq_scipy_style_args
        else:
            if capture_extra:
                diffeq_to_use = diffeq_args_extra
            else:
                diffeq_to_use = diffeq_args
    else:
        if use_scipy_style:
            if capture_extra:
                diffeq_to_use = diffeq_scipy_style_extra
            else:
                diffeq_to_use = diffeq_scipy_style
        else:
            if capture_extra:
                diffeq_to_use = diffeq_extra
            else:
                diffeq_to_use = diffeq

    if use_njit_always:
        diffeq_to_use = njit(diffeq_to_use)
    else:
        if (not njit_rk23_tested) and (integration_method=="RK23"):
            diffeq_to_use = njit(diffeq_to_use)
            njit_rk23_tested = True
        elif (not njit_rk45_tested) and (integration_method=="RK45"):
            diffeq_to_use = njit(diffeq_to_use)
            njit_rk45_tested = True
        elif (not njit_DOP853_tested) and (integration_method=="DOP853"):
            diffeq_to_use = njit(diffeq_to_use)
            njit_DOP853_tested = True

    if use_atol_array:
        atols_use = atols
    else:
        atols_use = atol
    if use_rtol_array:
        rtols_use = rtols
    else:
        rtols_use = rtol
    
    if use_different_tols:
        # Check that it can run with smaller tolerances
        atols_use = atols_use / 100.0
        rtols_use = rtols_use / 100.0
    
    if use_large_timespan:
        time_span_touse = time_span_large
    else:
        time_span_touse = time_span

    if use_args:
        args_touse = args
    else:
        args_touse = None
    
    if capture_extra:
        num_extra = 2
    else:
        num_extra = 0

    result = \
        pysolve_ivp(diffeq_to_use, time_span_touse, initial_conds,
                    method=integration_method,
                    args=args_touse, rtol=rtols_use, atol=atols_use,
                    num_extra=num_extra, first_step=first_step, max_step=max_step,
                    pass_dy_as_arg=(not use_scipy_style))

    assert isinstance(result, WrapCySolverResult)
    assert result.success
    assert result.error_code == 1
    assert result.size > 1
    assert result.message == "Integration completed without issue."
    # Check that the ndarrays make sense
    assert type(result.t) == np.ndarray
    assert result.t.dtype == np.float64
    assert result.y.dtype == np.float64
    assert result.t.size > 1
    assert result.t.size == result.y[0].size
    assert len(result.y.shape) == 2
    assert result.y[0].size == result.y[1].size
    assert result.t.size == result.size
    assert result.y[0].size == result.size

    if capture_extra:
        assert result.y.shape[0] == 4
        assert result.y[2].size == result.y[1].size
        assert result.y[3].size == result.y[1].size
    else:
        assert result.y.shape[0] == 2

    assert type(result.message) == str



def test_pysolve_ivp_errors():
    """ Test that the correct exceptions are raised when incorrect input is provided. """

    with pytest.raises(NotImplementedError) as e_info:
        # check for unsupported integrator
        result = pysolve_ivp(diffeq_scipy_style, time_span, initial_conds,
                        method="FakeIntegrationMethod",
                        args=None, rtol=rtol, atol=atol,
                        pass_dy_as_arg=False)
    
    with pytest.raises(AttributeError) as e_info:
        # Check for unsupported number of dependent variables
        result = pysolve_ivp(diffeq_scipy_style, time_span, np.linspace(0.0, 1.0, 1000),
                        method="RK23",
                        args=None, rtol=rtol, atol=atol,
                        pass_dy_as_arg=False)
    
    with pytest.raises(AttributeError) as e_info:
        # Check for unsupported number of additional captures
        result = pysolve_ivp(diffeq_scipy_style, time_span, initial_conds,
                        method="RK23",
                        args=None, rtol=rtol, atol=atol,
                        num_extra=1000,
                        pass_dy_as_arg=False)
    
    with pytest.raises(AttributeError) as e_info:
        # Check for bad first step
        result = pysolve_ivp(diffeq_scipy_style, time_span, initial_conds,
                        method="RK23",
                        args=None, rtol=rtol, atol=atol,
                        first_step = -10.0,
                        pass_dy_as_arg=False)
    
    with pytest.raises(AttributeError) as e_info:
        # Check for bad max step
        result = pysolve_ivp(diffeq_scipy_style, time_span, initial_conds,
                        method="RK23",
                        args=None, rtol=rtol, atol=atol,
                        max_step = -10.0,
                        pass_dy_as_arg=False)
        
    # Run integrator with a too small maximum number of steps. No exceptions should be raised but the integration should
    # not be successful
    result = pysolve_ivp(diffeq_scipy_style, time_span, initial_conds,
                        method="RK23",
                        args=None, rtol=rtol, atol=atol,
                        max_num_steps=5,
                        pass_dy_as_arg=False)

    assert not result.success
    assert result.error_code == -2
    assert result.message == "Maximum number of steps (set by user) exceeded during integration."

    # Do the same thing but now for max ram
    result = pysolve_ivp(diffeq_scipy_style, time_span, initial_conds,
                        method="RK23",
                        args=None, rtol=rtol, atol=atol,
                        max_ram_MB=0.0001,
                        pass_dy_as_arg=False)
    
    assert not result.success
    assert result.error_code == -3
    assert result.message == "Maximum number of steps (set by system architecture) exceeded during integration."

    # Do an integration with tolerances that are just way too small for the method
    result = pysolve_ivp(diffeq_stiff, time_span, initial_conds,
                        method="RK45",
                        args=None, rtol=1.0e-20, atol=1.0e-22,
                        pass_dy_as_arg=False)
    
    assert not result.success
    assert result.error_code == -1
    assert result.message == "Error in step size calculation:\n\tRequired step size is less than spacing between numbers."

@pytest.mark.parametrize('integration_method', ("RK23", "RK45", "DOP853"))
@pytest.mark.parametrize('t_eval_end', (None, 0.5, 1.0))
@pytest.mark.parametrize('test_dense_output', (False, True))
@pytest.mark.parametrize('backward_integrate', (False, True))
def test_pysolve_ivp_accuracy(integration_method, t_eval_end, test_dense_output, backward_integrate):
    #Check that the cython function solver is able to reproduce a known functions integral with reasonable accuracy

    # TODO: This is only checking one equation. Add other types of diffeqs to provide better coverage.

    # Differential Equation
    @njit
    def diffeq_accuracy(dy, t, y):
        dy[0] = np.sin(t) - y[1]  # dydt = sin(t) - x(t)
        dy[1] = np.cos(t) + y[0]  # dxdt = cos(t) + y(t)

    @njit
    def correct_answer(t, c1_, c2_):
        y = np.empty((2, t.size), dtype=np.float64)
        y[0] = -c1_ * np.sin(t) + c2_ * np.cos(t) - (np.cos(t) / 2)  # -c1 * sin(t) + c2 * cos(t) - cos(t) / 2
        # At t=0; y = c2 - 1/2
        y[1] = c2_ * np.sin(t) + c1_ * np.cos(t) + (np.sin(t) / 2)   # c2 * sin(t) + c1 * cos(t) + sin(t) / 2
        # At t=0; x = c1
        return y

    # Initial Conditions
    # y=0 --> c2 = + 1/2
    c2 = 0.5
    # x=1 --> c1 = + 1
    c1 = 1.0
    y0 = np.asarray((0., 1.), dtype=np.float64)
    time_span_ = (0., 10.)

    # Check if we should test t_eval
    if t_eval_end is None:
        t_eval = None
    else:
        t_eval = np.linspace(0.0, t_eval_end, 50)
    
    if backward_integrate:
        time_span_ = (time_span_[1], time_span_[0])
        if t_eval is not None:
            t_eval = np.ascontiguousarray(t_eval[::-1])

    result = \
        pysolve_ivp(diffeq_accuracy, time_span_, y0, method=integration_method, t_eval=t_eval, dense_output=test_dense_output,
                    rtol=1.0e-8, atol=1.0e-9, pass_dy_as_arg=True)
    
    # Use the integrator's time domain to build a correct solution
    real_answer = correct_answer(result.t, c1, c2)

    if t_eval is not None:
        assert result.t.size == t_eval.size
        assert np.allclose(result.t, t_eval)

    if integration_method == "RK23":
        check_rtol = 1.0e-3
        check_atol = 1.0e-6
    elif integration_method == "DOP853":
        check_rtol = 1.0e-5
        check_atol = 1.0e-8
    else:
        check_rtol = 1.0e-4
        check_atol = 1.0e-7
    
    if backward_integrate:
        pytest.skip("Backward integrating is not working well.")

    assert np.allclose(result.y, real_answer, rtol=check_rtol, atol=check_atol)

    if test_dense_output:
        # Check that dense output is working and that it gives decent results
        # Check with a float
        y_array = result(0.5)
        assert type(y_array) == np.ndarray
        assert y_array.shape == (2, 1)

        # Check with array
        t_array = np.linspace(0.1, 0.4, 10)
        y_array = result(t_array)
        assert type(y_array) == np.ndarray
        assert y_array.shape == (2, 10)

        # Check accuracy
        y_array_real = correct_answer(t_array, c1, c2)
        assert np.allclose(y_array, y_array_real, rtol=check_rtol, atol=check_atol)

    # Check the accuracy of the results
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # ax.plot(result.t, result.y[0], 'r', label='PySolver')
    # ax.plot(result.t, result.y[1], 'r:')
    # ax.plot(result.t, real_answer[0], 'b', label='Analytic')
    # ax.plot(result.t, real_answer[1], 'b:')
    # plt.show()

@pytest.mark.parametrize('integration_method', ("RK23", "RK45", "DOP853"))
def test_pysolve_ivp_readonly(integration_method):
    #Check that the cython function solver is able to reproduce a known functions integral with reasonable accuracy

    @njit
    def diffeq_accuracy(dy, t, y):
        dy[0] = np.sin(t) - y[1]  # dydt = sin(t) - x(t)
        dy[1] = np.cos(t) + y[0]  # dxdt = cos(t) + y(t)

    # Initial Conditions
    # y=0 --> c2 = + 1/2
    c2 = 0.5
    # x=1 --> c1 = + 1
    c1 = 1.0
    y0 = np.asarray((0., 1.), dtype=np.float64)
    # Make readonly
    y0.setflags(write=False)
    time_span_ = (0., 10.)

    result = \
        pysolve_ivp(diffeq_accuracy, time_span_, y0, method=integration_method,
                    rtol=1.0e-8, atol=1.0e-9, pass_dy_as_arg=True)
    
    assert isinstance(result, WrapCySolverResult)
    assert result.success
    assert result.error_code == 1
    assert result.size > 1
    assert result.message == "Integration completed without issue."
    # Check that the ndarrays make sense
    assert type(result.t) == np.ndarray
    assert result.t.dtype == np.float64
    assert result.y.dtype == np.float64
    assert result.t.size > 1
    assert result.t.size == result.y[0].size
    assert len(result.y.shape) == 2
    assert result.y[0].size == result.y[1].size
    assert result.t.size == result.size
    assert result.y[0].size == result.size
    assert result.y.shape[0] == 2
    assert type(result.message) == str



""" Lokavolterra Differential Equations """
import numpy as np
from numba import njit

from CyRK import nb2cy

lotkavolterra_y0 = np.asarray((10., 5.), dtype=np.float64)
lotkavolterra_args = (1.5, 1, 3, 1)
lotkavolterra_time_span_1 = (0., 15.)
lotkavolterra_time_span_2 = (0., 150.)


@njit(cache=True)
def lotkavolterra_nb(t, y, a, b, c, d):

    y0 = y[0]
    y1 = y[1]
    dy = np.empty_like(y)
    dy[0] = a * y0 - b * y0 * y1
    dy[1] = -c * y1 + d * y0 * y1
    return dy


lotkavolterra_cy = nb2cy(lotkavolterra_nb, use_njit=True, cache_njit=True)



def test_package():
    """Check if all the functions can be imported. """

    from CyRK import nbsolve_ivp, nb2cy, cy2nb, test_nbrk, test_cysolver, test_pysolver, version, __version__, pysolve_ivp

    assert type(version) == str
    assert version == __version__

def test_testers_nbrk():

    from CyRK import test_nbrk

    test_nbrk()

    assert True

def test_testers_cysolver():

    from CyRK import test_cysolver

    test_cysolver()

    assert True

def test_testers_pysolver():

    from CyRK import test_pysolver

    test_pysolver()

    assert True



from libcpp cimport bool as cpp_bool

from libc.math cimport sin, cos, fabs, fmin, fmax

from CyRK.cy.cysolver_api cimport cysolve_ivp, WrapCySolverResult, DiffeqFuncType,MAX_STEP, CySolveOutput, PreEvalFunc

cdef void baseline_diffeq(double* dy_ptr, double t, double* y_ptr, const void* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void accuracy_test_diffeq(double* dy_ptr, double t, double* y_ptr, const void* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void extraoutput_test_diffeq(double* dy_ptr, double t, double* y_ptr, const void* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void lorenz_diffeq(double* dy_ptr, double t, double* y_ptr, const void* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void lorenz_extraoutput_diffeq(double* dy_ptr, double t, double* y_ptr, const void* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void lotkavolterra_diffeq(double* dy_ptr, double t, double* y_ptr, const void* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void pendulum_diffeq(double* dy_ptr, double t, double* y_ptr, const void* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

from cython.parallel import prange

from libc.math cimport floor  

cdef void f() noexcept nogil:
    cdef int i, j, k, x
    for i in prange(100_000):
        for j in range(i):
            for k in range(j):
                x = i - j + <int>floor(k / 2)
    

f()
        


import os
import platform
import json
import sys
from setuptools import Extension, setup

import numpy as np

install_platform = platform.system()

if install_platform.lower() == 'windows':
    extra_compile_args = ['/openmp']
    extra_link_args = []
elif install_platform.lower() == 'darwin':
    # OpenMP is installed via llvm. See https://stackoverflow.com/questions/60005176/how-to-deal-with-clang-error-unsupported-option-fopenmp-on-travis
    extra_compile_args = ['-O3', '-fopenmp']
    extra_link_args = ['-lomp']
else:
    extra_compile_args = ['-fopenmp', '-O3']
    extra_link_args = ['-fopenmp', '-O3']
macro_list = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]

# Load CyRK's cython extensions
absolute_path = os.path.dirname(__file__)
cython_ext_path = os.path.join(absolute_path, 'cython_extensions.json')
with open(cython_ext_path, 'r') as cython_ext_file:
    cython_ext_dict = json.load(cython_ext_file)

cython_extensions = list()
for cython_ext, ext_data in cython_ext_dict.items():

    if ext_data['is_cpp']:
        if install_platform.lower() == 'windows':
            specific_compile_args = extra_compile_args + ext_data['compile_args'] + ["/std:c++20"]
        else:
            specific_compile_args = extra_compile_args + ext_data['compile_args'] + ["-std=c++20"]
    else:
        specific_compile_args = extra_compile_args + ext_data['compile_args']

    cython_extensions.append(
        Extension(
            name=ext_data['name'],
            sources=[os.path.join(*tuple(source_path)) for source_path in ext_data['sources']],
            # Always add numpy to any includes; also add sys.path so we can capture python.h
            include_dirs=[os.path.join(*tuple(dir_path)) for dir_path in ext_data['include_dirs']] + [np.get_include()] + sys.path,
            extra_compile_args=specific_compile_args,
            define_macros=macro_list,
            extra_link_args=ext_data['link_args'] + extra_link_args,
            )
        )

# Cython extensions require a setup.py in addition to pyproject.toml in order to create platform-specific wheels.
setup(
    ext_modules=cython_extensions,
    # cmdclass={'build_ext': build_ext, 'build_py':build_cyrk}
)


# distutils: language = c
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False
from libc.stdlib cimport malloc, realloc, free, exit, EXIT_FAILURE
from libc.stdio cimport printf


cdef inline void* allocate_mem(size_t size, char* var_name) noexcept nogil:
    cdef void* new_memory = malloc(size)
    if not new_memory:
        printf('Memory Error: Failed to allocate memory for %s \n\tRequested size = %zu.', var_name, size)
        exit(EXIT_FAILURE)
    return new_memory


cdef inline void* reallocate_mem(void* old_pointer, size_t new_size, char* var_name) noexcept nogil:
    cdef void* new_memory = realloc(old_pointer, new_size)
    if not new_memory:
        printf('Failed to *re*allocate memory for %s \n\tRequested size = %zu.', var_name, new_size)
        exit(EXIT_FAILURE)
    return new_memory

cdef inline void free_mem(void* pointer) noexcept nogil:
    free(pointer)



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False


cdef void interpolate_from_solution_list(
        double* y_result_ptr,
        int num_y,
        vector[CySolveOutput] solution_list_vec,
        int num_solutions,
        double* x_domain_ptr,
        size_t x_domain_size,
        vector[double] x_breakpoints_vec) noexcept nogil:
    """ Steps through multiple cysolve solutions to interpolate using a single all-domain domain. """
    
    cdef int current_sol_i = 0
    cdef double current_sol_upper_x = x_breakpoints_vec[0]
    cdef double* y_result_subptr = y_result_ptr
    cdef CySolverResult* cy_solution_ptr = solution_list_vec[0].get()

    cdef size_t x_i
    cdef double x
    
    for x_i in range(x_domain_size):
        
        x = x_domain_ptr[x_i]
        if x > current_sol_upper_x:
            # We are ready for the next solution.
            current_sol_i += 1
            current_sol_upper_x = x_breakpoints_vec[current_sol_i]
            cy_solution_ptr = solution_list_vec[current_sol_i].get()
        
        # Run cysolver interpolator
        y_result_subptr = &y_result_ptr[x_i * num_y]
        cy_solution_ptr.call(x, y_result_subptr)
